@@ -23,6 +23,7 @@
 #undef arma_align_mem
 #undef arma_warn_unused
 #undef arma_deprecated
+#undef arma_frown
 #undef arma_malloc
 #undef arma_inline
 #undef arma_noinline
@@ -34,6 +35,7 @@
 #define arma_align_mem
 #define arma_warn_unused
 #define arma_deprecated
+#define arma_frown(msg)
 #define arma_malloc
 #define arma_inline            inline
 #define arma_noinline
@@ -112,8 +114,10 @@
 
 
 #if defined(__APPLE__) || defined(__apple_build_version__)
-  #undef  ARMA_BLAS_SDOT_BUG
-  #define ARMA_BLAS_SDOT_BUG
+  // NOTE: Apple accelerate framework has broken implementations of functions that return a float value,
+  // NOTE: such as sdot(), slange(), clange(), slansy(), clanhe(), slangb(), snrm2(), sasum()
+  #undef  ARMA_BLAS_FLOAT_BUG
+  #define ARMA_BLAS_FLOAT_BUG
   
   // #undef  ARMA_HAVE_POSIX_MEMALIGN
   // NOTE: posix_memalign() is available since macOS 10.6 (late 2009 onwards)
@@ -138,11 +142,6 @@
 #endif
 
 
-// #if defined(ARMA_HAVE_CXX17)
-//   #define arma_warn_unused  [[nodiscard]]
-// #endif
-
-
 #if !defined(ARMA_ALLOW_FAKE_GCC)
   #if (defined(__GNUG__) || defined(__GNUC__)) && (defined(__INTEL_COMPILER) || defined(__NVCC__) || defined(__CUDACC__) || defined(__PGI) || defined(__PATHSCALE__) || defined(__ARMCC_VERSION) || defined(__IBMCPP__))
     #undef  ARMA_DETECTED_FAKE_GCC
@@ -162,10 +161,14 @@
   #undef  ARMA_GCC_VERSION
   #define ARMA_GCC_VERSION (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__)
   
-  #if (ARMA_GCC_VERSION < 40800)
-    #error "*** newer compiler required; need gcc 4.8 or later ***"
+  #if (ARMA_GCC_VERSION < 40803)
+    #error "*** newer compiler required; need gcc 4.8.3 or newer ***"
   #endif
   
+  // #if (ARMA_GCC_VERSION < 60100)
+  //   #pragma message ("WARNING: support for gcc versions older than 6.1 is deprecated")
+  // #endif
+  
   #define ARMA_GOOD_COMPILER
   
   #undef  arma_hot
@@ -174,19 +177,21 @@
   #undef  arma_align_mem
   #undef  arma_warn_unused
   #undef  arma_deprecated
+  #undef  arma_frown
   #undef  arma_malloc
   #undef  arma_inline
   #undef  arma_noinline
   
-  #define arma_hot                __attribute__((__hot__))
-  #define arma_cold               __attribute__((__cold__))
-  #define arma_aligned            __attribute__((__aligned__))
-  #define arma_align_mem          __attribute__((__aligned__(16)))
-  #define arma_warn_unused        __attribute__((__warn_unused_result__))
-  #define arma_deprecated         __attribute__((__deprecated__))
-  #define arma_malloc             __attribute__((__malloc__))
-  #define arma_inline      inline __attribute__((__always_inline__))
-  #define arma_noinline           __attribute__((__noinline__))
+  #define arma_hot         __attribute__((__hot__))
+  #define arma_cold        __attribute__((__cold__))
+  #define arma_aligned     __attribute__((__aligned__))
+  #define arma_align_mem   __attribute__((__aligned__(16)))
+  #define arma_warn_unused __attribute__((__warn_unused_result__))
+  #define arma_deprecated  __attribute__((__deprecated__))
+  #define arma_frown(msg)  __attribute__((__deprecated__(msg)))
+  #define arma_malloc      __attribute__((__malloc__))
+  #define arma_inline      __attribute__((__always_inline__)) inline
+  #define arma_noinline    __attribute__((__noinline__))
   
   #undef  ARMA_HAVE_ALIGNED_ATTRIBUTE
   #define ARMA_HAVE_ALIGNED_ATTRIBUTE
@@ -204,6 +209,7 @@
 #endif
 
 
+// TODO: __INTEL_CLANG_COMPILER indicates the clang based intel compiler, distinct from the classic intel compiler
 #if !defined(ARMA_ALLOW_FAKE_CLANG)
   #if defined(__clang__) && (defined(__INTEL_COMPILER) || defined(__NVCC__) || defined(__CUDACC__) || defined(__PGI) || defined(__PATHSCALE__) || defined(__ARMCC_VERSION) || defined(__IBMCPP__))
     #undef  ARMA_DETECTED_FAKE_CLANG
@@ -247,6 +253,11 @@
     #define arma_deprecated __attribute__((__deprecated__))
   #endif
   
+  #if __has_attribute(__deprecated__)
+    #undef  arma_frown
+    #define arma_frown(msg) __attribute__((__deprecated__(msg)))
+  #endif
+  
   #if __has_attribute(__malloc__)
     #undef  arma_malloc
     #define arma_malloc __attribute__((__malloc__))
@@ -254,7 +265,7 @@
   
   #if __has_attribute(__always_inline__)
     #undef  arma_inline
-    #define arma_inline inline __attribute__((__always_inline__))
+    #define arma_inline __attribute__((__always_inline__)) inline
   #endif
   
   #if __has_attribute(__noinline__)
@@ -309,7 +320,7 @@
   #undef  arma_deprecated
   #define arma_deprecated __declspec(deprecated)
   // #undef  arma_inline
-  // #define arma_inline inline __forceinline
+  // #define arma_inline __forceinline inline
   
   #pragma warning(push)
   
@@ -321,6 +332,7 @@
   #pragma warning(disable: 4512)  // assignment operator can't be generated
   #pragma warning(disable: 4513)  // destructor can't be generated
   #pragma warning(disable: 4514)  // unreferenced inline function has been removed
+  #pragma warning(disable: 4519)  // default template args are only allowed on a class template (C++11)
   #pragma warning(disable: 4522)  // multiple assignment operators specified
   #pragma warning(disable: 4623)  // default constructor can't be generated
   #pragma warning(disable: 4624)  // destructor can't be generated
@@ -331,7 +343,8 @@
   #pragma warning(disable: 4711)  // call was inlined
   #pragma warning(disable: 4714)  // __forceinline can't be inlined
   #pragma warning(disable: 4800)  // value forced to bool
-  #pragma warning(disable: 4519)  // C++11: default template args are only allowed on a class template
+  
+  // NOTE: also possible to disable 4146 (unary minus operator applied to unsigned type, result still unsigned)
   
   #if defined(ARMA_HAVE_CXX17)
   #pragma warning(disable: 26812)  // unscoped enum
@@ -364,19 +377,26 @@
 
 
 #if defined(__SUNPRO_CC)
-  
   // http://www.oracle.com/technetwork/server-storage/solarisstudio/training/index-jsp-141991.html
   // http://www.oracle.com/technetwork/server-storage/solarisstudio/documentation/cplusplus-faq-355066.html
-  
   #if (__SUNPRO_CC < 0x5140)
     #error "*** newer compiler required ***"
   #endif
-  
 #endif
 
 
-#if defined(__CYGWIN__) && !defined(ARMA_DONT_PRINT_CXX11_WARNING)
-  #pragma message ("WARNING: Cygwin may have incomplete support for C++11 features.")
+#if defined(ARMA_HAVE_CXX14)
+  #undef  arma_deprecated
+  #define arma_deprecated [[deprecated]]
+
+  #undef  arma_frown
+  #define arma_frown(msg) [[deprecated(msg)]]
+#endif
+
+
+#if defined(ARMA_HAVE_CXX17)
+  #undef  arma_warn_unused
+  #define arma_warn_unused  [[nodiscard]]
 #endif
 
 
@@ -401,7 +421,7 @@
   #pragma message ("WARNING: use of OpenMP disabled; compiler support for OpenMP 3.1+ not detected")
   
   #if (defined(_OPENMP) && (_OPENMP < 201107))
-    #pragma message ("NOTE: your compiler appears to have an ancient version of OpenMP")
+    #pragma message ("NOTE: your compiler has an outdated version of OpenMP")
     #pragma message ("NOTE: consider upgrading to a better compiler")
   #endif
 #endif
@@ -410,7 +430,6 @@
 #if defined(ARMA_USE_OPENMP)
   #if (defined(ARMA_GCC_VERSION) && (ARMA_GCC_VERSION < 50400))
     // due to https://gcc.gnu.org/bugzilla/show_bug.cgi?id=57580
-    // TODO: gcc 4.9.4 is also fixed, so use a more fine-grained gcc version check?
     #undef ARMA_USE_OPENMP
     #if !defined(ARMA_DONT_PRINT_OPENMP_WARNING)
       #pragma message ("WARNING: use of OpenMP disabled due to compiler bug in gcc <= 5.3")
@@ -419,6 +438,19 @@
 #endif
 
 
+#if (defined(__FAST_MATH__) || (defined(__FINITE_MATH_ONLY__) && (__FINITE_MATH_ONLY__ > 0)) || defined(_M_FP_FAST))
+  #undef  ARMA_FAST_MATH
+  #define ARMA_FAST_MATH
+#endif
+
+
+#if defined(ARMA_FAST_MATH) && !defined(ARMA_DONT_PRINT_FAST_MATH_WARNING)
+  #pragma message ("WARNING: compiler is in fast math mode; some functions may be unreliable.")
+  #pragma message ("WARNING: to suppress this warning and related warnings,")
+  #pragma message ("WARNING: #define ARMA_DONT_PRINT_FAST_MATH_WARNING before #include <armadillo>")
+#endif
+
+
 #if ( (defined(_WIN32) || defined(_WIN64) || defined(_MSC_VER)) && (!defined(__MINGW32__) && !defined(__MINGW64__)) )
   #undef  ARMA_PRINT_EXCEPTIONS_INTERNAL
   #define ARMA_PRINT_EXCEPTIONS_INTERNAL
@@ -440,36 +472,37 @@
 
 
 
+// undefine conflicting macros
+
 #if defined(log2)
   #undef log2
-  #pragma message ("WARNING: detected 'log2' macro and undefined it")
+  #pragma message ("WARNING: undefined conflicting 'log2' macro")
 #endif
 
-
-
-// 
-// whoever defined macros with the names "min" and "max" should be permanently removed from the gene pool
+#if defined(check)
+  #undef check
+  #pragma message ("WARNING: undefined conflicting 'check' macro")
+#endif
 
 #if defined(min) || defined(max)
   #undef min
   #undef max
-  #pragma message ("WARNING: detected 'min' and/or 'max' macros and undefined them;")
-  #pragma message ("WARNING: you may wish to define NOMINMAX before including any windows header")
+  #pragma message ("WARNING: undefined conflicting 'min' and/or 'max' macros")
 #endif
 
-
-
-//
-// handle more stupid macros
 // https://sourceware.org/bugzilla/show_bug.cgi?id=19239
-
 #undef minor
 #undef major
 
 
 // optionally allow disabling of compile-time deprecation messages (not recommended)
+// NOTE: option 'ARMA_IGNORE_DEPRECATED_MARKER' will be removed
+// NOTE: disabling deprecation messages is counter-productive
 
 #if defined(ARMA_IGNORE_DEPRECATED_MARKER) && (!defined(ARMA_DONT_IGNORE_DEPRECATED_MARKER)) && (!defined(ARMA_EXTRA_DEBUG))
   #undef  arma_deprecated
   #define arma_deprecated
+
+  #undef  arma_frown
+  #define arma_frown(msg)
 #endif
